/*
 * Copyright (C) 1997 Gary Howland (gary@hotlava.com)
 * All rights reserved.
 */

package hotlava.net.http;

import java.io.*;
import java.net.Socket;
import java.net.InetAddress;


/**
 * A class that extends Thread, that is used to
 * "mop up" left over incoming data on sockets
 *
 * @version $Revision: 1.1 $
 *
 * <p>Copyright (C) 1997 Gary Howland (gary@hotlava.com)
 * All rights reserved.
 */
final class ConnectionClearer extends Thread
{
	/**
	 * The InputStream being cleared
	 */
	InputStream in;

	/**
	 * The PrintStream where debugging is sent, unless null
	 */
	private PrintStream debug;


	/**
	 * Creates a new connection clearer thread
	 *
	 * @param in the InputStream to clear
	 * @param debug the PrintStream where debugging is sent, unless null
	 */
	ConnectionClearer(InputStream in, PrintStream debug)
	{
		this.in = in;
		this.debug = debug;
		setPriority(getPriority()-2);
	}
	
	/**
	 * Read all the data from the InputStream being cleared,
	 * and close it.
	 */
	public void
	run()
	{
		try {
			while (true)
				if (in.read() == -1)
					break;
		}
		catch (IOException e)
		{
			if ("recv".equals(e.getMessage()))
				return;
			if (debug != null)
			{
				debug.println("Unexpected exception in connection clearer");
				e.printStackTrace(debug);
			}
		}
		finally { try { in.close(); } catch (IOException e) {} }
	}
}


/**
 * A class that extends Thread, that is created by Server
 * for every incoming connection.
 *
 * @see Server
 * @version $version$
 *
 * <p>Copyright (C) 1997 Gary Howland (gary@hotlava.com)
 * All rights reserved.
 */
public final class Connection
	extends Thread
{
	/**
	 * The PrintStream where debugging is sent, unless null
	 */
	private PrintStream debug;

	/**
	 * The Request handler for this connection
	 */
	private Handler handler;

	/**
	 * The socket for this connection
	 */
	private Socket s;

	/**
	 * The address originating this connection
	 */
	private InetAddress addr;

	/**
	 * True if all data must be read from the socket before closing
	 */
	boolean clearInput;
	
	/**
	 * True if requests should have a trailing \r\n
	 */
	boolean trailingCrNl;
	
	/**
	 * Set the debug PrintStream
	 * @param out the PrintStream where debugging is sent, unless null
	 */
	public void debug(PrintStream out) { debug = out; }

	/**
	 * Define whether all data must be read from the socket before closing
	 * @param flag true if all data should be read from the socket
	 */
	public void clearInput(boolean flag) { clearInput = flag; }

	/**
	 * Define whether returned data has a trailing \r\n, like Netscape
	 * @param flag true if \r\n should have a trailing \r\n
	 */
	public void trailingCrNl(boolean flag) { trailingCrNl = flag; }


	/**
	 * Creates a Connection object
	 *
	 * @param handler the Handler for this connection
	 * @param s the (accepted) incoming connection socket
	 */
	Connection(Handler handler, Socket s)
	{
		this.s = s;
		this.handler = handler;

		this.trailingCrNl = true;

		// setPriority(getPriority()+2);
	}

	/**
	 * Reads a Request from the socket, and then passes it to the Handler.
	 * The return value from the Handler will be sent back out on the socket.
	 */
	public void
	run()
	{
		boolean response_sent = false;

		InetAddress addr = s.getInetAddress();

		if (debug != null)
			debug.println("Accepted connection from "+addr.getHostName());


		InputStream is = null;
		try {
			is = new BufferedInputStream(s.getInputStream());
			DataInputStream dis = new DataInputStream(is);
			Request req = new Request(dis);

			//
			//	Clearinput hack
			//
			if (clearInput)
			{
				if (debug != null)
					debug.println("Clearing data in new thread");

				new ConnectionClearer(dis, debug).start();
			}


			//
			//	Process request
			//
			BufferedOutputStream bos = new BufferedOutputStream(s.getOutputStream());
			DataOutputStream dos = new DataOutputStream(bos);

			try {
				handler.handleRequest(req, dos);

				//
				// Emulate Netscape browsers ...
				//
				if (trailingCrNl)
				{
					dos.writeByte(13);
					dos.writeByte(10);
				}

				dos.flush();
				dos.close();

				response_sent = true;
			}
			catch (HandlerException e) {
				if (debug != null)
					debug.println("HTTP Request exception: "+e.getMessage());
			}
			//
			// TODO: I think there may be EOF exceptions here.  Gary.
			//
			catch (IOException e) {
				//
				// This could have failed if the browser is no longer listening
				// e.g. if the user interrupted the loading of the page
				// so perhaps we should catch the IOExceptions?
				// (which will be of the type "Broken pipe")
				//
				String msg = e.getMessage();
				if (msg != null && msg.equals("Broken pipe"))
				{
					if (debug != null)
						debug.println("Browser closed connection");
					response_sent = true; // Don't bother sending any data
				}
				else if (msg != null && msg.equals("Permission denied"))
				{
					if (debug != null)
						debug.println("Could not connect");
				}
				else
					throw e;
			}
		}
		catch (MalformedRequestException e)
		{
			if (debug != null)
				debug.println("Malformed request: "+e.getMessage());
		}
		catch (IOException e)
		{
			if (debug != null)
				e.printStackTrace(debug);
		}
		finally
		{
			if (response_sent == false)
			{
				try {
					PrintStream out = new PrintStream(s.getOutputStream());
					out.print("HTTP/1.0 204 No response\r\n\r\n");
				}
				catch (IOException e) {
					if (debug != null)
						e.printStackTrace(debug);
				}
			}

			if (is != null)
				try {
					if (clearInput)
						is.close();
				}
				catch (IOException e) {
					if (debug != null)
						e.printStackTrace(debug);
				}

			if (s != null)
				try {
					s.close();
				}
				catch (IOException e) {
					if (debug != null)
						e.printStackTrace(debug);
				}

		}
	}

}
