/*
 * Copyright (C) 1997 Gary Howland (gary@hotlava.com)
 * All rights reserved.
 */
 
package hotlava.net.http;

import java.io.*;
import java.net.*;

/**
 * This class implements Handler in order to implement
 * a proxying Handler.
 * Objects of this class can themselves be configured
 * to use a proxy.
 *
 * TODO: This class needs some work to keep connections alive,
 * in order to improve performance.
 *
 * @see Handler
 * @version $Revision: 1.1 $
 *
 * <p>Copyright (C) 1997 Gary Howland (gary@hotlava.com)
 * All rights reserved.
 */
public class ProxyHandler implements Handler
{
	/**
	 *	Stream to which all debugging goes, unless null
	 */
	private PrintStream debug;

    /**
	 * Set the debug PrintStream
	 * @param out the PrintStream where debugging is sent, unless null
	 */
	public void debug(PrintStream out) { debug = out; }

	/**
	 *	Proxy host, or null if no proxy required.
	 */
	private String proxy_host = null;

	/**
	 *	Proxy port
	 */
	private int proxy_port = 8080;


	/**
	 * Creates a ProxyHandler object, which implements the Handler interface.
	 * This class is used to proxy HTTP Requests.
	 */
	public ProxyHandler() {}

	/**
	 * Creates a ProxyHandler object, which is implements the Handler interface.
	 * This class is used to proxy HTTP Requests.
	 *
	 * @param proxy_host the proxy host
	 * @param proxy_port the proxy port
	 */
	public
	ProxyHandler(String proxy_host, int proxy_port)
	{
		this.proxy_host = proxy_host;
		this.proxy_port = proxy_port;
	}

	/**
	 * Set the host used for proxying requests
	 * @param host The proxy host, or null if no proxy required
	 */
	public void
	host(String host)
	{
		proxy_host = host;
	}

	/**
	 * Set the port used for proxying requests
	 * @param port The proxy port
	 */
	public void
	port(int port)
	{
		proxy_port = port;
	}

	/**  
	 * Handle a request
	 *
	 * @param r the Request to be handled
	 * @param os the DataOutput stream where the result of the request is sent
	 * @exception IOException If an I/O error has occurred.
	 * @exception HandlerException If a the Request could not be handled.
	 */     
	public void
	handleRequest(Request r, DataOutput os)
		throws IOException, HandlerException
	{
		if (debug != null)
		{
			debug.println("Received request: [" + r.toString() + "]");
			debug.println("Method = "+r.method());
			debug.println("Host = "+r.host());
			debug.println("Port = "+r.port());
			debug.println("URI = "+r.URI());
			debug.println("Headers = "+r.header_data());
		}

		if (r.host() == null)
			throw new HandlerException("Host required for proxy request");

		Socket s = null;
		if (proxy_host == null)
		{
			if (debug != null)
				debug.println("Connecting to "+r.host()+" port "+r.port());
			s = new Socket(r.host(), r.port());
		}
		else
		{
			if (debug != null)
				debug.println("Connecting to "+proxy_host+" port "+proxy_port);
			s = new Socket(proxy_host, proxy_port);
		}

		PrintStream out = new PrintStream(s.getOutputStream());


		//
		// Build the request line
		//
		StringBuffer reqbuf = new StringBuffer();
		reqbuf.append(r.method()).append(" ");

		if (proxy_host != null)
		{
			reqbuf.append("http://");
			reqbuf.append(r.host());
			if (r.port() != 80)
				reqbuf.append(":").append(String.valueOf(r.port()));
		}

		reqbuf.append(r.abs_path());

		String version = r.version();
		if (version != null)
			reqbuf.append(" ").append(version);

		reqbuf.append("\r\n");
		reqbuf.append(r.header_data());
		reqbuf.append("\r\n");
		
		byte[] data = r.data();
		if (data != null && data.length > 0)
		{
			ByteArrayOutputStream buf = new ByteArrayOutputStream();
			new DataOutputStream(buf).write(data);
			reqbuf.append(buf.toString());
		}

		String req = reqbuf.toString();
		out.print(req);
		out.flush();

		if (debug != null)
			debug.println("Sending request: [" + req + "]");

		InputStream in = s.getInputStream();
		byte buf[] = new byte[8192]; // create new tmp buffer.
		while (true) // will change to add big buffer.
		{
			int len = in.read(buf);
			if (len == -1)
				break;

			//
			// It would be nice to do a hex dump here
			//

			os.write(buf, 0, len);
		}
	}
}
