/*
 * Copyright (C) 1997 Gary Howland (gary@hotlava.com)
 * All rights reserved.
 */

package hotlava.net.http;

import java.io.*;
import java.util.*;

/**
 * This class represents an HTTP Reply
 *
 * @version $Revision: 1.1 $
 *
 * <p>Copyright (C) 1997 Gary Howland (gary@hotlava.com)
 * All rights reserved.
 */
public class Reply
{
	protected String header_data;
	protected Properties headers;
	protected byte[] data;
	
	/**
	 * Get a reply header field
	 * @param key the field name as a String
	 * @return the field value as a String
	 */
	public String getHeader(String key) { return headers.getProperty(key); }

	/**
	 * Get a reply header field, but specify a default
	 * @param key the field name as a String
	 * @param def the default value to use if the field is not present
	 * @return the field value as a String
	 */
	public String getHeader(String key, String def) { return headers.getProperty(key, def); }

	/**
	 * Set a reply header field
	 * @param key the field name as a String
	 * @param value the value to be assigned to the field
	 */
	public void setHeader(String key, String value) { headers.put(key, value); }

	/**
	 * Remove a reply header field
	 * @param key the field name (as a String) to be removed
	 * @return the value of the removed header
	 */
	public String removeHeader(String key) { return (String)headers.remove(key); }

	/**
	 * Return the message body
	 * @return a byte array containing the message body
	 */
	public byte[] data() { return data; }

	/**
	 * Define the message body
	 * @param data a byte array containing the message body
	 */
	public void
	data(byte[] data)
	{
		this.data = data;
		setHeader("content-length", String.valueOf(data.length));
	}

	/**
	 * Creates a new Reply from an InputStream
	 * @param in InputStream
	 * @exception IOException an IO error has occurred
	 */
	public
	Reply(InputStream in)
		throws IOException, MalformedReplyException
	{
		DataInputStream dis = new DataInputStream(in);

		//
		//	Read the headers
		//
		headers = new Properties();
		while (true)
		{
			String hline = dis.readLine();
			if (hline == null)
				throw new MalformedReplyException("Unexpected EOF");
			if (hline.length() == 0)
				break;

			header_data += hline + "\r\n";

			//
			// TODO - If this line starts with a space or tab,
			// it is a continuation of the previous line
			//
			int pos;
			if ((pos = hline.indexOf(":")) != -1)
			{
				String key = hline.substring(0, pos).toLowerCase();

				// TODO - this assumes one space after the :
				// there can be any amount of white space there
				String value = hline.substring(pos+1).trim();


				//
				// Multiple fields are always comma separed lists
				// (4.2 of the Http/1.1 spec)
				//
				String s = headers.getProperty(key);
				if (s != null)
					headers.put(key, s + ", " + value);
				else
					headers.put(key, value);
			}
		}

		//
		//	Read any data sent from the user agent
		//
		String contentLength = headers.getProperty("content-length");
		if (contentLength != null)
		{
			int len;
			try {
				len = Integer.parseInt(contentLength);
			}
			catch(NumberFormatException e) {
				throw new MalformedReplyException("Content-Length not numeric");
			}
			dis.readFully(data = new byte[len]);
		}
	}

	/**
	 * Returns a String object containing all of the reply header field
	 */
	public String
	header_data()
	{
		StringBuffer r = new StringBuffer();

		for (Enumeration e = headers.propertyNames(); e.hasMoreElements() ;)
		{
			String s = (String)e.nextElement();
			r.append(s).append(": ").append(headers.getProperty(s)).append("\r\n");
		}
		return r.toString();
	}

	/**
	 * Returns a String object containing most of the reply details
	 */
	public String
	toString()
	{
		return "[Reply headers: " + headers.toString() +"]";
	}
}
