/*
 * Copyright (C) 1997 Gary Howland (gary@hotlava.com)
 * All rights reserved.
 */

package hotlava.net.http;

import java.io.*;
import java.util.*;
import java.net.MalformedURLException;

import java.net.URLEncoder;	// Sun only provide an encoder, not a decoder :-/
import hotlava.net.URLDecoder;

/**
 * This class represents an HTTP Request
 *
 * @version $Revision: 1.1 $
 *
 * <p>Copyright (C) 1997 Gary Howland (gary@hotlava.com)
 * All rights reserved.
 */
public class Request
{
	private String request_line;
	private String method;
	private String protocol = null;
	private String request_uri;
	private String http_version = null;
	private String filename;
	private String params;
	private Properties parameters = null;
	private Properties headers;
	private byte[] data;

	private String http_host = null;
	private int http_port = -1;
	private String abs_path = "";
	private String header_data = "";
	
	/**
	 * Returns the initial request line as a string
	 */
	public String request_line() { return request_line; }

	/**
	 * Returns the request method as a string
	 */
	public String method() { return method; }

	/**
	 * Returns the URI as a string
	 */
	public String
	URI()
	{
		String r = "";

		if (protocol != null)
			r += protocol + "://";

		if (http_host != null)
			r += http_host;

		if (http_port != -1)
			r += ":" + String.valueOf(http_port);

		r += abs_path();

		return r;
	}

	/**
	 * Returns the request version as a string
	 */
	public String version() { return (http_version == null) ? "HTTP/0.9" : http_version; }

	/**
	 * Returns the URI host component as a string
	 */
	public String host() { return http_host; }

	/**
	 * Set the URI host component
	 */
	public void host(String host)
	{
		http_host = host;
		setHeader("host", host);
	}

	/**
	 * Returns the URI port component as a string
	 */
	public int port() { return (http_port == -1) ? 80 : http_port; }

	/**
	 * Set the URI port component
	 */
	public void port(int port) { http_port = port; }

	/**
	 * Returns the URI absolute path component as a string
	 */
	public String
	abs_path()
	{
		String r = "";

		r += filename;

		String p = params();
		if (p != "")
			r += "?" + p;

		return r;
	}

	/**
	 * Returns the URI parameters as a string
	 */
	public String
	params()
	{
		if (parameters == null)
			parseParams();

		String r = "";
		for (Enumeration e = parameters.propertyNames(); e.hasMoreElements() ;)
		{
			String s = (String)e.nextElement();
			r += URLEncoder.encode(s) + "=" + URLEncoder.encode(parameters.getProperty(s));
			if (e.hasMoreElements())
				r += "&";
		}
		return r;
	}

	/**
	 * Returns the value of a parameter
	 * @param key the name of the parameter
	 * @return A String object containing the value, or null if parameter not present
	 */
	public String
	getParam(String key)
	{
		if (parameters == null)
			parseParams();

		return parameters.getProperty(key);
	}

	/**
	 * Sets the value of a parameter
	 * @param key the name of the parameter
	 * @param value the value to assign to the parameter
	 */
	public void
	setParam(String key, String value)
	{
		if (parameters == null)
			parseParams();

		parameters.put(key, value);
	}

	/**
	 * Remove a parameter
	 * @param key the name of the parameter to be removed
	 * @return the value of the removed parameter
	 */
	public String
	removeParam(String key)
	{
		if (parameters == null)
			parseParams();

		return (String)parameters.remove(key);
	}

	/**
	 * Parse the params string into a Properties object.
	 * Only does this if it has not already been done.
	 * The parsing is provided as a function so that it can be
	 * delayed for efficiency.
	 */
	protected void
	parseParams()
	{
		if (parameters != null)
			return;

		parameters = new Properties();

		int last = 0, pos;
		while ((pos = params.indexOf('&', last)) != -1)
		{
			if (last != pos)
				parseParamPair(params.substring(last, pos));
			last = pos + 1;
		}
		parseParamPair(params.substring(last));
	}

	private void
	parseParamPair(String s)
	{
		int pos = s.indexOf('=');

		if (pos == -1) // Treat parameters with an '=' as empty
			parameters.put(s, "");
		else
			parameters.put(s.substring(0, pos), s.substring(pos+1));
	}


	/**
	 * Returns the URI filename component as a string
	 */
	public String filename() { return filename; }

	/**
	 * Sets the URI filename component
	 */
	public void filename(String filename) { this.filename = filename; }

	/**
	 * Sets the URI abs_path component
	 */
	public void abs_path(String abs_path) { this.abs_path = abs_path; }



	/**
	 * Sets the request method
	 * @param method the method as a String
	 */
	public void method(String method) { this.method = method; }

	/**
	 * Sets the URI
	 * @param request_uri the URI as a String
	 */
	public void
	URI(String request_uri)
		throws MalformedRequestException
	{
		int pos;

		params = ""; parameters = null;	// Remove existing parameters
		abs_path = request_uri;	// This may be overriden if we have an http:// URL

		//
		// If http URL
		//
		if (request_uri.startsWith("http://"))
		{
			protocol = "http";

			String t = request_uri.substring(7);
			if ((pos = t.indexOf('/')) != -1)
			{
				abs_path = t.substring(pos);
				t = t.substring(0, pos);
			}

			if ((pos = t.indexOf(':')) == -1)
			{
				http_host = t;
			}
			else
			{
				http_host = t.substring(0, pos);
				try {
					http_port = Integer.parseInt(t.substring(pos+1));
				}
				catch(NumberFormatException e) {
					throw new MalformedRequestException("HTTP URL port not numeric ("+t.substring(pos+1)+")");
				}
			}
		}

		//
		// Strip off parameters, if any
		//
		pos = abs_path.indexOf('?');
		if (pos != -1)
		{
			filename = abs_path.substring(0, pos);
			pos++;
			if (abs_path.length() > pos)
				params = abs_path.substring(pos);
			else
				params = "";
		}
		else
		{
			filename = abs_path;
			params = "";
		}
	}


	/**
	 * Get a request header field
	 * @param key the field name as a String
	 * @return the field value as a String
	 */
	public String getHeader(String key) { return headers.getProperty(key); }

	/**
	 * Get a request header field, but specify a default
	 * @param key the field name as a String
	 * @param def the default value to use if the field is not present
	 * @return the field value as a String
	 */
	public String getHeader(String key, String def) { return headers.getProperty(key, def); }

	/**
	 * Set a request header field
	 * @param key the field name as a String
	 * @param value the value to be assigned to the field
	 */
	public void setHeader(String key, String value) { headers.put(key, value); }

	/**
	 * Remove a request header field
	 * @param key the field name (as a String) to be removed
	 * @return the value of the removed header
	 */
	public String removeHeader(String key) { return (String)headers.remove(key); }


	/**
	 * Convert the parameters into postable data
	 * Also set the method type to POST,
	 * and the content-type to application/x-www-form-urlencoded
	 */
	public void
	convertParamsForPosting()
	{
		try {
			ByteArrayOutputStream buf = new ByteArrayOutputStream();
			new DataOutputStream(buf).writeBytes(params());
			data(buf.toByteArray());
		}
		catch (IOException e) {
			throw new RuntimeException("Unexpected error: "+e.getMessage());
		}
		params = ""; parameters = null;
		setHeader("content-type", "application/x-www-form-urlencoded");
		method("POST");
	}

	/**
	 * Return the message body
	 * @return a byte array containing the message body
	 */
	public byte[] data() { return data; }

	/**
	 * Define the message body
	 * @param data a byte array containing the message body
	 */
	public void
	data(byte[] data)
	{
		this.data = data;
		setHeader("content-length", String.valueOf(data.length));
	}

	/**
	 * Creates a new Request from a DataInput stream
	 * @param in A DataInput stream
	 * @exception IOException an IO error has occurred
	 * @exception MalformedRequestException The request is badly formatted
	 */
	public
	Request(DataInput in)
		throws IOException, MalformedRequestException
	{
		request_line = in.readLine();
		if (request_line == null)
			throw new MalformedRequestException("Null Request-Line");

		int pos;
		if ((pos = request_line.indexOf(' ')) == -1)
			throw new MalformedRequestException("Invalid Request-Line: " + request_line);

		method = request_line.substring(0, pos);

		pos++;
		String request_uri = request_line.substring(pos);

		if ((pos = request_uri.indexOf(' ')) == -1)
		{
			//
			//	We have a Simple-Request
			//
		}
		else
		{
			String t = request_uri.substring(0, pos);
			pos++;
			http_version = request_uri.substring(pos);
			request_uri = t;

			if ((!http_version.equals("HTTP/1.0")) && (!http_version.equals("HTTP/1.1")))
				throw new MalformedRequestException("Invalid HTTP-Version: " + http_version);
		}

		//
		//	Unescape the URI
		//
		try {
			request_uri = URLDecoder.decode(request_uri);
		}
		catch (MalformedURLException e) {
			throw new MalformedRequestException("Invalid URL escaping: "+e.getMessage());
		}

		//
		// Parse the URI and set the appropriate variables
		//
		URI(request_uri);


		//
		//	Read the message headers
		//
		headers = new Properties();
		while (true)
		{
			String hline = in.readLine();
			if (hline == null)
				throw new MalformedRequestException("Unexpected EOF");
			if (hline.length() == 0)
				break;

			header_data += hline + "\r\n";

			//
			// TODO - If this line starts with a space or tab,
			// it is a continuation of the previous line
			//

			if ((pos = hline.indexOf(":")) != -1)
			{
				String key = hline.substring(0, pos).toLowerCase();

				// TODO - this assumes one space after the :
				// there can be any amount of white space there
				String value = hline.substring(pos+1).trim();


				//
				// Multiple fields are always comma separed lists
				// (4.2 of the Http/1.1 spec)
				//
				String s = headers.getProperty(key);
				if (s != null)
					headers.put(key, s + ", " + value);
				else
					headers.put(key, value);
			}
		}

		//
		//	Read any data sent from the user agent
		//
		String contentLength = headers.getProperty("content-length");
		if (contentLength != null)
		{
			int len;
			try {
				len = Integer.parseInt(contentLength);
			}
			catch(NumberFormatException e) {
				throw new MalformedRequestException("Content-Length not numeric");
			}
			in.readFully(data = new byte[len]);
		}


		//
		// Now test for POSTed parameters
		//
		if (!method.equals("POST"))
			return;

		String ct = headers.getProperty("content-type");
		if (ct == null || (!ct.equals("application/x-www-form-urlencoded")))
			return;

		//
		//	Unescape the parameters
		//
		try {
			data = URLDecoder.decode(data);
		}
		catch (MalformedURLException e) {
			throw new MalformedRequestException("Invalid URL escaping: "+e.getMessage());
		}

		parameters = null;
		params = new String(data, 0);

	}

	/**
	 * Returns a String object containing all of the request header field
	 */
	public String
	header_data()
	{
		StringBuffer r = new StringBuffer();

		for (Enumeration e = headers.propertyNames(); e.hasMoreElements() ;)
		{
			String s = (String)e.nextElement();
			r.append(s).append(": ").append(headers.getProperty(s)).append("\r\n");
		}
		return r.toString();
	}

	/**
	 * Returns a String object containing most of the request details
	 */
	public String
	toString()
	{
		return "[Method: " + method + ","
			+ "URI: " + URI() + ","
			+ "HTTP Version: " + version() + ","
			+ "Message headers: " + headers.toString() +"]";
	}
}
