/*
 * Copyright (C) 1997 Gary Howland (gary@hotlava.com)
 * All rights reserved.
 */

package hotlava.net.http;

import java.io.*;
import java.net.Socket;
import java.net.ServerSocket;


/**
 * This class listens on a port, creates a new Connection object
 * for every incoming connection, which in turn reads in a Request
 * from the socket, and passes the Request to a Handler.
 *
 * @see Connection
 * @see Handler
 * @see Request
 * @version $Revision: 1.1 $
 *
 * <p>Copyright (C) 1997 Gary Howland (gary@hotlava.com)
 * All rights reserved.
 */
public class Server
{
	/**
	 * The port which this server listens on
	 */
	private int port;

	/**
	 * The PrintStream where debugging is sent, unless null
	 */
	private PrintStream debug;

	/**
	 * True if all data must be read from the socket before closing
	 */
	private boolean clearInput;

	/**
	 * The Handler object that will deal with Requests
	 */
	private Handler handler;


	/**
	 * Set the port which the server listens on
	 * @param port the port number
	 */
	public void port(int port) { this.port = port; }

	/**
	 * Set the debug PrintStream
	 * @param out the PrintStream where debugging is sent, unless null
	 */
	public void debug(PrintStream out) { debug = out; }

	/**
	 * Define whether all data must be read from the socket before closing
	 * @param flag true if all data should be read from the socket
	 */
	public void clearInput(boolean flag) { clearInput = flag; }

	/**
	 * Set the handler object that will deal with Requests
	 * @param handler a Handler object to deal with Requests
	 */
	public void handler(Handler handler) { this.handler = handler; }


	/**
	 * Creates a Server object
	 *
	 * @param poer the port that this server will listen on
	 */
	public
	Server(int port)
	{
		this.port = port;
	}

	/**
	 * Start serving
	 * @exception IOException an IO error has occurred
	 */
	public void
	serve()
		throws IOException
	{
		ServerSocket listen = null;
		try {
			listen = new ServerSocket(port);

			if (debug != null)
				debug.println("Starting server on port "+port);

			while (true)
			{
				Socket s = listen.accept();

				if (debug != null)
					debug.println("Incoming connection accepted");

				Connection sc = new Connection(handler, s);
				sc.debug(debug);
				sc.clearInput(clearInput);
				sc.start();
			}

		}
		finally {
			if (debug != null)
				debug.println("Closing socket");

			if (listen != null)
				listen.close();
		}
	}


	/**
	 * A test server
	 */
	public static void
	main(String[] argv)
		throws IOException
	{
		ProxyHandler proxy = new ProxyHandler();
		proxy.debug(System.out);

		Server s = new Server(8083);
		s.debug(System.out);
		s.handler(proxy);

		s.serve();
	}
}
