/*
 * Copyright (C) 1997 Gary Howland (gary@hotlava.com)
 * All rights reserved.
 */

package hotlava.util;

import java.util.Hashtable;
import java.util.Enumeration;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.DataInputStream;
import java.io.PrintStream;
import java.io.IOException;

// Needed for testing
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;

/**
 * This class reads and writes windows style(ish) ini files
 * It has a multi line extension using the unix 'here' format:
 * If this entry is used: <pre>item = << END<pre><p>
 * then all the lines up until the line <pre>END<pre> will be
 * read in as one line.
 *
 * All  relevant methods are synchronized, so that an IniFile object
 * can be shared between many threads.
 *
 * Should this class be rewritten using StreamTokenizer?
 *
 * @author Hotlava Ltd
 * @version $Revision: 1.1 $
 * @version $State: Exp $
 *
 * <p>Copyright (C) 1997 Gary Howland (gary@hotlava.com)
 * All rights reserved.
 */

public class IniFile
{
	protected Hashtable data = new Hashtable();

	public IniFile() {}

	/**
	 * Creates an IniFile object from an InputStream
	 * @exception IOException
	 */
	public IniFile(InputStream in)
		throws IOException
	{
		read(in);
	}

	/**
	 * Updates an IniFile object from an InputStream
	 * @exception IOException
	 */
	public synchronized void 
	read(InputStream raw_in)
		throws IOException
	{
		DataInputStream in = ( raw_in instanceof DataInputStream )
				? (DataInputStream)raw_in : new DataInputStream( raw_in );
		Hashtable h = null, data = this.data = new Hashtable();
		String s, end = null, item = null;
		boolean multiline = false;
		StringBuffer sb = null;

		int lineno = 0;
		String section = null;

	read:
		while ((s = in.readLine()) != null)
		{
			lineno++;

			if ( !multiline )
			{
				int i;
				if ( ( s = s.trim() ).length() == 0 )
					continue;
				switch ( s.charAt( 0 ) )
				{
				case '[':
					if ((i = s.indexOf( "]")) == - 1)
						throw new IniFileFormatException(section, lineno, "Invalid section definition: " + s);
					section = s.substring(1, i);
					data.put(section, h = new Hashtable());
					continue read;

				// comments
				case ';':
				case '#':
					continue read;

				default:
					if (section == null)
						throw new IniFileFormatException(section, lineno, "Missing section definition");
					if ((i = s.indexOf( "=" )) == - 1)
						throw new IniFileFormatException(section, lineno, "Invalid definition: " + s);
					item  = s.substring( 0, i ).trim();
					s = s.substring( i + 1 ).trim();
					if ( !s.startsWith( "<<" ) )
					{
						h.put( item, s );
						continue read;
					}
					if ((end = s.substring( 2 ).trim()).length() == 0) 
						throw new IniFileFormatException(section, lineno, "Missing end definition");
					multiline = true;
					sb = null;
					continue read;
				}
			}
			// multi line
			if (end.equals(s.trim()))
			{
				if (section == null)
					throw new IniFileFormatException(section, lineno, "Missing section definition");
				h.put( item, sb.toString() );
				multiline = false;
			}

			if (sb == null)
				sb = new StringBuffer();
			else
				sb.append('\n');
			sb.append(s);
		}
	}
	
	/**
	 * Save an IniFile object to an OutputStream
	 */
	public synchronized void 
	save(OutputStream raw_out)
	{
		boolean first_section = true;

		PrintStream out = ( raw_out instanceof PrintStream )
				? (PrintStream)raw_out : new PrintStream( raw_out );
		Hashtable data = this.data;
		for (Enumeration e = data.keys(); e.hasMoreElements(); )
		{
			String section;
			Hashtable h = (Hashtable)data.get( section = (String)e.nextElement() );
			if (!first_section)
				out.println();
			first_section = false;
			out.println( "[" + section + "]" );
			for ( Enumeration items = h.keys(); items.hasMoreElements(); )
			{
				String item;
				String value = (String)h.get( item = (String)items.nextElement() );
				if ( value.indexOf( "\n" ) == -1 ) // single line
				{
					out.println("    "+ item + "=" + value );
					continue;
				}
				out.println("    "+ item + "=<<END\n" + value + "\nEND" );
			}
		}
	}
	
	/**
	 * Get a value from the IniFile
	 * @param section the section from which to find the field
	 * @param item the field from which to read a value
	 * @return a String object with the field contents, or null if field not present
	 */
	public synchronized String
	get(String section, String item)
	{
		return get( section, item, null );
	}

	/**
	 * Get the value from the IniFile
	 * @param section the section from which to find the field
	 * @param item the field from which to read a value
	 * @param def a default value to use if the field is not present
	 * @return a String object with the field contents, or the default
	 */
	public synchronized String
	get( String section, String item, String def )
	{
		Hashtable h;
		if ( ( h = (Hashtable)data.get( section ) ) != null )
		{
			String s;
			if ( ( s = (String)h.get( item ) ) != null )
				return s;
		}
		return def;
	}

	/**
	 * Set a value in the IniFile
	 * @param section the section in which to find the field
	 * @param item the field which is being set
	 * @param value the value to set the field to
	 * @return a String object with the previous field contents, or null if none
	 */
	public synchronized String
	set( String section, String item, String value )
	{
		Hashtable h;
		if ( ( h = (Hashtable)data.get( section ) ) == null )
			data.put( section, h = new Hashtable() );
		return (String)h.put( item, value );
	}
	
	/**
	 * Determine if a section exists
	 * @param section the section being queried
	 * @return true if the section exists, false otherwise
	 */
	public synchronized boolean
	sectionExists( String section )
	{
		return data.get( section ) != null;
	}

	/**
	 * Obtain all of the section headings
	 * @return an enumeration containing all the section headings
	 */
	public Enumeration
	sections()
	{
		return data.keys();
	}

	/**
	 * Obtain all of the field names for a particular section
	 * @param section the section being queried
	 * @return an enumeration containing all the field names, or null if the section doesn't exit
	 */
	public Enumeration
	items( String section )
	{
		Hashtable h;
		if ( ( h = (Hashtable)data.get( section ) ) != null )
			return h.keys();
		return null;
	}



	//
	//  Test code
	//
	public static void
	main(String argv[])
	throws IOException
	{
		File in = new File(".", argv[0]);
		System.out.println("Reading file "+argv[0]);

		FileInputStream fis = new FileInputStream(in);

		IniFile ini = new IniFile(fis);

		System.out.println("[local] local_name = " + ini.get("local", "local_name"));
		System.out.println("[local] local_issuer_url = " + ini.get("local", "local_issuer_url"));

		System.out.println( "Dump in hash order ..." );
		for ( Enumeration e = ini.sections(); e.hasMoreElements(); )
		{
			String section = e.nextElement().toString();
			System.out.println( "[" + section + "]" );
			for ( Enumeration i = ini.items( section ); i.hasMoreElements(); )
			{
				String item = i.nextElement().toString();
				System.out.println( item + "=" + ini.get( section, item, "<NULL>" ) );
			}
		}
		System.out.println( "<EOF>" );
	}
}

